"use strict";

var __importStar = this && this.__importStar || function (mod) {
  if (mod && mod.__esModule) return mod;
  var result = {};
  if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
  result["default"] = mod;
  return result;
};

Object.defineProperty(exports, "__esModule", {
  value: true
});

var bt = __importStar(require("@babel/types"));

var bhelpers = __importStar(require("ts2odata-babel"));

var traverse_1 = require("ts2odata-babel");

var EntitySetContext_1 = require("ts2odata");

var PluginVisitor =
/** @class */
function () {
  function PluginVisitor() {
    this.ArrowFunctionExpression = function (path, state) {
      replaceFunctionNode(path, state);
    };

    this.CallExpression = function (path, state) {
      var _a;

      if (bt.isMemberExpression(path.node.callee) && bt.isIdentifier(path.node.callee.property)) {
        if (path.node.callee.property.name === 'getQueryUrl' || path.node.callee.property.name === 'toArrayAsync') {
          var queryPaser = new QueryParser(getQueries(state), path.scope, (_a = state.opts) === null || _a === void 0 ? void 0 : _a.odataNamespace);
          queryPaser.parse(path.node);
        }
      }
    };

    this.FunctionExpression = function (path, state) {
      replaceFunctionNode(path, state);
    };

    this.Program = function (path, state) {
      var moduleName = state.opts.checkModuleName;
      var isFoundModuleName = fixModuleDeclaration(path.node, moduleName);
      if (moduleName && !isFoundModuleName) path.stop();
    };
  }

  return PluginVisitor;
}();

exports.PluginVisitor = PluginVisitor;

var QueryParser =
/** @class */
function () {
  function QueryParser(queries, scope, odataNamespace) {
    this.queries = queries;
    this.scope = scope;
    this.babelTraverse = new traverse_1.Traverse();
    this.entitySetContext = new EntitySetContext_1.EntitySetContext('http://dummy', 'dummy', null, '', odataNamespace);
    this.nodes = new Array();
  }

  QueryParser.prototype.getScope = function (node) {
    if (node.arguments.length > 1) {
      if (bt.isIdentifier(node.arguments[1])) return this.getScopeByName(node.arguments[1].name);
      if (bt.isObjectExpression(node.arguments[1])) return this.getScopeFromObjectExpression(node.arguments[1]);
      throw new Error('Unknown scope argument type ' + node.arguments[1].type);
    }

    return undefined;
  };

  QueryParser.prototype.getScopeByName = function (scopeName) {
    var binding = this.scope.getBinding(scopeName);
    if (binding && bt.isVariableDeclarator(binding.path.node) && bt.isObjectExpression(binding.path.node.init)) return this.getScopeFromObjectExpression(binding.path.node.init);
    throw new Error('Variable ' + scopeName + ' not found');
  };

  QueryParser.prototype.getScopeFromObjectExpression = function (node) {
    var scope = {};

    for (var i = 0; i < node.properties.length; i++) {
      var property = node.properties[i];
      if (bt.isObjectProperty(property) && bt.isIdentifier(property.key)) Object.defineProperty(scope, property.key.name, {
        get: function () {
          return '';
        }
      });
    }

    return scope;
  };

  QueryParser.prototype.parse = function (node) {
    if (bt.isMemberExpression(node.callee) && bt.isIdentifier(node.callee.property)) {
      this.nodes.push(node);
      if (bt.isCallExpression(node.callee.object)) this.parse(node.callee.object);
    }

    this.parseNode();
  };

  QueryParser.prototype.parseNode = function () {
    while (this.nodes.length > 0) {
      var node = this.nodes.pop();

      if (node.arguments.length > 2) {
        this.queries.clear();
        return;
      }

      var callee = node.callee;
      var property = callee.property;

      switch (property.name) {
        case 'expand':
        case 'orderby':
        case 'orderbyDescending':
        case 'thenExpand':
        case 'thenOrderby':
        case 'thenOrderbyDescending':
          this.traversePropertyPath(node);
          break;

        case 'filter':
        case 'thenFilter':
          this.traverseFilter(node);
          break;

        case 'groupby':
          this.traverseGroupby(node);
          break;

        case 'key':
          this.traverseKey(node);
          break;

        case 'select':
        case 'thenSelect':
          this.traverseSelect(node);
          break;

        case 'asEntitySet':
        case 'getQueryUrl':
        case 'skip':
        case 'thenSkip':
        case 'thenTop':
        case 'toArrayAsync':
        case 'top':
          break;

        default:
          this.queries.clear();
          return;
      }
    }
  };

  QueryParser.prototype.traverseFilter = function (node) {
    if (bhelpers.isFunctionExpression(node.arguments[0])) {
      var query = this.babelTraverse.traverseAstFilter(this.entitySetContext, node.arguments[0], this.getScope(node));
      this.queries.set(node.arguments[0], query);
    }
  };

  QueryParser.prototype.traverseGroupby = function (node) {
    var properties = this.traverseSelect(node);

    for (var i = 0; i < properties.length; i++) this.entitySetContext.addGroupBy(properties[i]);
  };

  QueryParser.prototype.traverseKey = function (node) {
    if (node.arguments.length > 1 && bhelpers.isFunctionExpression(node.arguments[1])) this.traversePropertyPathNode(node.arguments[1]);
  };

  QueryParser.prototype.traversePropertyPath = function (node) {
    if (bhelpers.isFunctionExpression(node.arguments[0])) this.traversePropertyPathNode(node.arguments[0]);
  };

  QueryParser.prototype.traversePropertyPathNode = function (node) {
    var body = bhelpers.getFunctionBody(node);
    var expression = this.babelTraverse.traverseAstPropertyPath(body);
    if (expression === undefined) throw new Error('Invlid property path');
    this.queries.set(node, expression);
  };

  QueryParser.prototype.traverseSelect = function (node) {
    if (bhelpers.isFunctionExpression(node.arguments[0])) {
      var queries = this.babelTraverse.traverseAstSelect(this.entitySetContext, node.arguments[0], this.getScope(node));
      var json = JSON.stringify(queries);
      this.queries.set(node.arguments[0], json);
      return queries;
    }

    return [];
  };

  return QueryParser;
}();

function fixModuleDeclaration(node, moduleName) {
  var isFoundModuleName = false;

  for (var _i = 0, _a = node.body; _i < _a.length; _i++) {
    var statement = _a[_i];
    if (bt.isVariableDeclaration(statement)) for (var _b = 0, _c = statement.declarations; _b < _c.length; _b++) {
      var declaration = _c[_b];

      if (bt.isCallExpression(declaration.init)) {
        var requireNode = declaration.init;
        if (bt.isIdentifier(requireNode.callee) && requireNode.callee.name === '__importStar') if (requireNode.arguments.length === 1 && bt.isCallExpression(requireNode.arguments[0])) requireNode = requireNode.arguments[0];
        if (bt.isIdentifier(requireNode.callee) && requireNode.callee.name === 'require') if (requireNode.arguments.length === 1 && bt.isStringLiteral(requireNode.arguments[0])) {
          if (requireNode.arguments[0].value === moduleName) isFoundModuleName = true;
          if (requireNode.arguments[0].value === 'ts2odata-babel') requireNode.arguments[0].value = 'ts2odata';
        }
      }
    }
  }

  return isFoundModuleName;
}

function getQueries(state) {
  if (!state.queries) state.queries = new Map();
  return state.queries;
}

function replaceFunctionNode(path, state) {
  var query = getQueries(state).get(path.node);
  if (query) path.replaceWith(bt.stringLiteral(query));
}